from random import shuffle

import pytest

import numpy as np
from numpy.testing import assert_allclose
from numpy.testing import assert_array_equal

from skimage.transform import integral_image
from skimage.feature import haar_like_feature
from skimage.feature import haar_like_feature_coord
from skimage.feature import draw_haar_like_feature


def test_haar_like_feature_error():
    img = np.ones((5, 5), dtype=np.float32)
    img_ii = integral_image(img)

    feature_type = 'unknown_type'
    with pytest.raises(ValueError):
        haar_like_feature(img_ii, 0, 0, 5, 5, feature_type=feature_type)
        haar_like_feature_coord(5, 5, feature_type=feature_type)
        draw_haar_like_feature(img, 0, 0, 5, 5, feature_type=feature_type)

    feat_coord, feat_type = haar_like_feature_coord(5, 5, 'type-2-x')
    with pytest.raises(ValueError):
        haar_like_feature(
            img_ii, 0, 0, 5, 5, feature_type=feat_type[:3], feature_coord=feat_coord
        )


@pytest.mark.parametrize("dtype", [np.uint8, np.int8, np.float32, np.float64])
@pytest.mark.parametrize(
    "feature_type,shape_feature,expected_feature_value",
    [
        ('type-2-x', (84,), [0.0]),
        ('type-2-y', (84,), [0.0]),
        ('type-3-x', (42,), [-5, -4.0, -3.0, -2.0, -1.0]),
        ('type-3-y', (42,), [-5, -4.0, -3.0, -2.0, -1.0]),
        ('type-4', (36,), [0.0]),
    ],
)
def test_haar_like_feature(feature_type, shape_feature, expected_feature_value, dtype):
    # test Haar-like feature on a basic one image
    img = np.ones((5, 5), dtype=dtype)
    img_ii = integral_image(img)
    haar_feature = haar_like_feature(img_ii, 0, 0, 5, 5, feature_type=feature_type)
    assert_allclose(np.sort(np.unique(haar_feature)), expected_feature_value)


@pytest.mark.parametrize("dtype", [np.uint8, np.int8, np.float32, np.float64])
@pytest.mark.parametrize(
    "feature_type", ['type-2-x', 'type-2-y', 'type-3-x', 'type-3-y', 'type-4']
)
def test_haar_like_feature_fused_type(dtype, feature_type):
    # check that the input type is kept
    img = np.ones((5, 5), dtype=dtype)
    img_ii = integral_image(img)
    expected_dtype = img_ii.dtype
    # to avoid overflow, unsigned type are converted to signed
    if 'uint' in expected_dtype.name:
        expected_dtype = np.dtype(expected_dtype.name.replace('u', ''))
    haar_feature = haar_like_feature(img_ii, 0, 0, 5, 5, feature_type=feature_type)
    assert haar_feature.dtype == expected_dtype


def test_haar_like_feature_list():
    img = np.ones((5, 5), dtype=np.int8)
    img_ii = integral_image(img)
    feature_type = ['type-2-x', 'type-2-y', 'type-3-x', 'type-3-y', 'type-4']
    haar_list = haar_like_feature(img_ii, 0, 0, 5, 5, feature_type=feature_type)
    haar_all = haar_like_feature(img_ii, 0, 0, 5, 5)
    assert_array_equal(haar_list, haar_all)


@pytest.mark.parametrize(
    "feature_type",
    [
        'type-2-x',
        'type-2-y',
        'type-3-x',
        'type-3-y',
        'type-4',
        ['type-2-y', 'type-3-x', 'type-4'],
    ],
)
def test_haar_like_feature_precomputed(feature_type):
    img = np.ones((5, 5), dtype=np.int8)
    img_ii = integral_image(img)
    if isinstance(feature_type, list):
        # shuffle the index of the feature to be sure that we are output
        # the features in the same order
        shuffle(feature_type)
        feat_coord, feat_type = zip(
            *[haar_like_feature_coord(5, 5, feat_t) for feat_t in feature_type]
        )
        feat_coord = np.concatenate(feat_coord)
        feat_type = np.concatenate(feat_type)
    else:
        feat_coord, feat_type = haar_like_feature_coord(5, 5, feature_type)
    haar_feature_precomputed = haar_like_feature(
        img_ii, 0, 0, 5, 5, feature_type=feat_type, feature_coord=feat_coord
    )
    haar_feature = haar_like_feature(img_ii, 0, 0, 5, 5, feature_type)
    assert_array_equal(haar_feature_precomputed, haar_feature)


@pytest.mark.parametrize(
    "feature_type,height,width,expected_coord",
    [
        (
            'type-2-x',
            2,
            2,
            [
                [[(0, 0), (0, 0)], [(0, 1), (0, 1)]],
                [[(0, 0), (1, 0)], [(0, 1), (1, 1)]],
                [[(1, 0), (1, 0)], [(1, 1), (1, 1)]],
            ],
        ),
        (
            'type-2-y',
            2,
            2,
            [
                [[(0, 0), (0, 0)], [(1, 0), (1, 0)]],
                [[(0, 0), (0, 1)], [(1, 0), (1, 1)]],
                [[(0, 1), (0, 1)], [(1, 1), (1, 1)]],
            ],
        ),
        (
            'type-3-x',
            3,
            3,
            [
                [[(0, 0), (0, 0)], [(0, 1), (0, 1)], [(0, 2), (0, 2)]],
                [[(0, 0), (1, 0)], [(0, 1), (1, 1)], [(0, 2), (1, 2)]],
                [[(0, 0), (2, 0)], [(0, 1), (2, 1)], [(0, 2), (2, 2)]],
                [[(1, 0), (1, 0)], [(1, 1), (1, 1)], [(1, 2), (1, 2)]],
                [[(1, 0), (2, 0)], [(1, 1), (2, 1)], [(1, 2), (2, 2)]],
                [[(2, 0), (2, 0)], [(2, 1), (2, 1)], [(2, 2), (2, 2)]],
            ],
        ),
        (
            'type-3-y',
            3,
            3,
            [
                [[(0, 0), (0, 0)], [(1, 0), (1, 0)], [(2, 0), (2, 0)]],
                [[(0, 0), (0, 1)], [(1, 0), (1, 1)], [(2, 0), (2, 1)]],
                [[(0, 0), (0, 2)], [(1, 0), (1, 2)], [(2, 0), (2, 2)]],
                [[(0, 1), (0, 1)], [(1, 1), (1, 1)], [(2, 1), (2, 1)]],
                [[(0, 1), (0, 2)], [(1, 1), (1, 2)], [(2, 1), (2, 2)]],
                [[(0, 2), (0, 2)], [(1, 2), (1, 2)], [(2, 2), (2, 2)]],
            ],
        ),
        (
            'type-4',
            2,
            2,
            [[[(0, 0), (0, 0)], [(0, 1), (0, 1)], [(1, 1), (1, 1)], [(1, 0), (1, 0)]]],
        ),
    ],
)
def test_haar_like_feature_coord(feature_type, height, width, expected_coord):
    feat_coord, feat_type = haar_like_feature_coord(width, height, feature_type)
    # convert the output to a full numpy array just for comparison
    feat_coord = np.array([hf for hf in feat_coord])
    assert_array_equal(feat_coord, expected_coord)
    assert np.all(feat_type == feature_type)


@pytest.mark.parametrize("max_n_features,nnz_values", [(None, 46), (1, 4)])
def test_draw_haar_like_feature(max_n_features, nnz_values):
    img = np.zeros((5, 5), dtype=np.float32)
    coord, _ = haar_like_feature_coord(5, 5, 'type-4')
    image = draw_haar_like_feature(
        img, 0, 0, 5, 5, coord, max_n_features=max_n_features, rng=0
    )
    draw_haar_like_feature(img, 0, 0, 5, 5, coord, max_n_features=max_n_features, rng=0)
    assert image.shape == (5, 5, 3)
    assert np.count_nonzero(image) == nnz_values



import itertools

import numpy as np
import pytest

from skimage._shared._dependency_checks import has_mpl
from skimage._shared._warnings import expected_warnings
from skimage._shared.testing import run_in_parallel
from skimage._shared.utils import _supported_float_type, convert_to_float
from skimage.data import shepp_logan_phantom
from skimage.transform import radon, iradon, iradon_sart, rescale


PHANTOM = shepp_logan_phantom()[::2, ::2]
PHANTOM = rescale(
    PHANTOM, 0.5, order=1, mode='constant', anti_aliasing=False, channel_axis=None
)


def _debug_plot(original, result, sinogram=None):
    from matplotlib import pyplot as plt

    imkwargs = dict(cmap='gray', interpolation='nearest')
    if sinogram is None:
        plt.figure(figsize=(15, 6))
        sp = 130
    else:
        plt.figure(figsize=(11, 11))
        sp = 221
        plt.subplot(sp + 0)
        plt.imshow(sinogram, aspect='auto', **imkwargs)
    plt.subplot(sp + 1)
    plt.imshow(original, **imkwargs)
    plt.subplot(sp + 2)
    plt.imshow(result, vmin=original.min(), vmax=original.max(), **imkwargs)
    plt.subplot(sp + 3)
    plt.imshow(result - original, **imkwargs)
    plt.colorbar()
    plt.show()


def _rescale_intensity(x):
    x = x.astype(float)
    x -= x.min()
    x /= x.max()
    return x


def test_iradon_bias_circular_phantom():
    """
    test that a uniform circular phantom has a small reconstruction bias
    """
    pixels = 128
    xy = np.arange(-pixels / 2, pixels / 2) + 0.5
    x, y = np.meshgrid(xy, xy)
    image = x**2 + y**2 <= (pixels / 4) ** 2

    theta = np.linspace(0.0, 180.0, max(image.shape), endpoint=False)
    sinogram = radon(image, theta=theta)

    reconstruction_fbp = iradon(sinogram, theta=theta)
    error = reconstruction_fbp - image

    tol = 5e-5
    roi_err = np.abs(np.mean(error))
    assert roi_err < tol


def check_radon_center(shape, circle, dtype, preserve_range):
    # Create a test image with only a single non-zero pixel at the origin
    image = np.zeros(shape, dtype=dtype)
    image[(shape[0] // 2, shape[1] // 2)] = 1.0
    # Calculate the sinogram
    theta = np.linspace(0.0, 180.0, max(shape), endpoint=False)
    sinogram = radon(image, theta=theta, circle=circle, preserve_range=preserve_range)
    assert sinogram.dtype == _supported_float_type(sinogram.dtype)
    # The sinogram should be a straight, horizontal line
    sinogram_max = np.argmax(sinogram, axis=0)
    print(sinogram_max)
    assert np.std(sinogram_max) < 1e-6


@pytest.mark.parametrize("shape", [(16, 16), (17, 17)])
@pytest.mark.parametrize("circle", [False, True])
@pytest.mark.parametrize("dtype", [np.float64, np.float32, np.float16, np.uint8, bool])
@pytest.mark.parametrize("preserve_range", [False, True])
def test_radon_center(shape, circle, dtype, preserve_range):
    check_radon_center(shape, circle, dtype, preserve_range)


@pytest.mark.parametrize("shape", [(32, 16), (33, 17)])
@pytest.mark.parametrize("circle", [False])
@pytest.mark.parametrize("dtype", [np.float64, np.float32, np.uint8, bool])
@pytest.mark.parametrize("preserve_range", [False, True])
def test_radon_center_rectangular(shape, circle, dtype, preserve_range):
    check_radon_center(shape, circle, dtype, preserve_range)


def check_iradon_center(size, theta, circle):
    debug = False
    # Create a test sinogram corresponding to a single projection
    # with a single non-zero pixel at the rotation center
    if circle:
        sinogram = np.zeros((size, 1), dtype=float)
        sinogram[size // 2, 0] = 1.0
    else:
        diagonal = int(np.ceil(np.sqrt(2) * size))
        sinogram = np.zeros((diagonal, 1), dtype=float)
        sinogram[sinogram.shape[0] // 2, 0] = 1.0
    maxpoint = np.unravel_index(np.argmax(sinogram), sinogram.shape)
    print('shape of generated sinogram', sinogram.shape)
    print('maximum in generated sinogram', maxpoint)
    # Compare reconstructions for theta=angle and theta=angle + 180;
    # these should be exactly equal
    reconstruction = iradon(sinogram, theta=[theta], circle=circle)
    reconstruction_opposite = iradon(sinogram, theta=[theta + 180], circle=circle)
    print(
        'rms deviance:',
        np.sqrt(np.mean((reconstruction_opposite - reconstruction) ** 2)),
    )
    if debug and has_mpl:
        import matplotlib.pyplot as plt

        imkwargs = dict(cmap='gray', interpolation='nearest')
        plt.figure()
        plt.subplot(221)
        plt.imshow(sinogram, **imkwargs)
        plt.subplot(222)
        plt.imshow(reconstruction_opposite - reconstruction, **imkwargs)
        plt.subplot(223)
        plt.imshow(reconstruction, **imkwargs)
        plt.subplot(224)
        plt.imshow(reconstruction_opposite, **imkwargs)
        plt.show()

    assert np.allclose(reconstruction, reconstruction_opposite)


sizes_for_test_iradon_center = [16, 17]
thetas_for_test_iradon_center = [0, 90]
circles_for_test_iradon_center = [False, True]


@pytest.mark.parametrize(
    "size, theta, circle",
    itertools.product(
        sizes_for_test_iradon_center,
        thetas_for_test_iradon_center,
        circles_for_test_iradon_center,
    ),
)
def test_iradon_center(size, theta, circle):
    check_iradon_center(size, theta, circle)


def check_radon_iradon(interpolation_type, filter_type):
    debug = False
    image = PHANTOM
    reconstructed = iradon(
        radon(image, circle=False),
        filter_name=filter_type,
        interpolation=interpolation_type,
        circle=False,
    )
    delta = np.mean(np.abs(image - reconstructed))
    print('\n\tmean error:', delta)
    if debug and has_mpl:
        _debug_plot(image, reconstructed)
    if filter_type in ('ramp', 'shepp-logan'):
        if interpolation_type == 'nearest':
            allowed_delta = 0.03
        else:
            allowed_delta = 0.025
    else:
        allowed_delta = 0.05
    assert delta < allowed_delta


filter_types = ["ramp", "shepp-logan", "cosine", "hamming", "hann"]
interpolation_types = ['linear', 'nearest']
radon_iradon_inputs = list(itertools.product(interpolation_types, filter_types))
# cubic interpolation is slow; only run one test for it
radon_iradon_inputs.append(('cubic', 'shepp-logan'))


@pytest.mark.parametrize("interpolation_type, filter_type", radon_iradon_inputs)
def test_radon_iradon(interpolation_type, filter_type):
    check_radon_iradon(interpolation_type, filter_type)


def test_iradon_angles():
    """
    Test with different number of projections
    """
    size = 100
    # Synthetic data
    image = np.tri(size) + np.tri(size)[::-1]
    # Large number of projections: a good quality is expected
    nb_angles = 200
    theta = np.linspace(0, 180, nb_angles, endpoint=False)
    radon_image_200 = radon(image, theta=theta, circle=False)
    reconstructed = iradon(radon_image_200, circle=False)
    delta_200 = np.mean(
        abs(_rescale_intensity(image) - _rescale_intensity(reconstructed))
    )
    assert delta_200 < 0.03
    # Lower number of projections
    nb_angles = 80
    radon_image_80 = radon(image, theta=theta, circle=False)
    # Test whether the sum of all projections is approximately the same
    s = radon_image_80.sum(axis=0)
    assert np.allclose(s, s[0], rtol=0.01)
    reconstructed = iradon(radon_image_80, circle=False)
    delta_80 = np.mean(
        abs(image / np.max(image) - reconstructed / np.max(reconstructed))
    )
    # Loss of quality when the number of projections is reduced
    assert delta_80 > delta_200


def check_radon_iradon_minimal(shape, slices):
    debug = False
    theta = np.arange(180)
    image = np.zeros(shape, dtype=float)
    image[slices] = 1.0
    sinogram = radon(image, theta, circle=False)
    reconstructed = iradon(sinogram, theta, circle=False)
    print('\n\tMaximum deviation:', np.max(np.abs(image - reconstructed)))
    if debug and has_mpl:
        _debug_plot(image, reconstructed, sinogram)
    if image.sum() == 1:
        assert np.unravel_index(
            np.argmax(reconstructed), image.shape
        ) == np.unravel_index(np.argmax(image), image.shape)


shapes = [(3, 3), (4, 4), (5, 5)]


def generate_test_data_for_radon_iradon_minimal(shapes):
    def shape2coordinates(shape):
        c0, c1 = shape[0] // 2, shape[1] // 2
        coordinates = itertools.product((c0 - 1, c0, c0 + 1), (c1 - 1, c1, c1 + 1))
        return coordinates

    def shape2shapeandcoordinates(shape):
        return itertools.product([shape], shape2coordinates(shape))

    return itertools.chain.from_iterable(
        [shape2shapeandcoordinates(shape) for shape in shapes]
    )


@pytest.mark.parametrize(
    "shape, coordinate", generate_test_data_for_radon_iradon_minimal(shapes)
)
def test_radon_iradon_minimal(shape, coordinate):
    check_radon_iradon_minimal(shape, coordinate)


def test_reconstruct_with_wrong_angles():
    a = np.zeros((3, 3))
    p = radon(a, theta=[0, 1, 2], circle=False)
    iradon(p, theta=[0, 1, 2], circle=False)
    with pytest.raises(ValueError):
        iradon(p, theta=[0, 1, 2, 3])


def _random_circle(shape):
    # Synthetic random data, zero outside reconstruction circle
    np.random.seed(98312871)
    image = np.random.rand(*shape)
    c0, c1 = np.ogrid[0 : shape[0], 0 : shape[1]]
    r = np.sqrt((c0 - shape[0] // 2) ** 2 + (c1 - shape[1] // 2) ** 2)
    radius = min(shape) // 2
    image[r > radius] = 0.0
    return image


def test_radon_circle():
    a = np.ones((10, 10))
    with expected_warnings(['reconstruction circle']):
        radon(a, circle=True)

    # Synthetic data, circular symmetry
    shape = (61, 79)
    c0, c1 = np.ogrid[0 : shape[0], 0 : shape[1]]
    r = np.sqrt((c0 - shape[0] // 2) ** 2 + (c1 - shape[1] // 2) ** 2)
    radius = min(shape) // 2
    image = np.clip(radius - r, 0, np.inf)
    image = _rescale_intensity(image)
    angles = np.linspace(0, 180, min(shape), endpoint=False)
    sinogram = radon(image, theta=angles, circle=True)
    assert np.all(sinogram.std(axis=1) < 1e-2)

    # Synthetic data, random
    image = _random_circle(shape)
    sinogram = radon(image, theta=angles, circle=True)
    mass = sinogram.sum(axis=0)
    average_mass = mass.mean()
    relative_error = np.abs(mass - average_mass) / average_mass
    print(relative_error.max(), relative_error.mean())
    assert np.all(relative_error < 3.2e-3)


def check_sinogram_circle_to_square(size):
    from skimage.transform.radon_transform import _sinogram_circle_to_square

    image = _random_circle((size, size))
    theta = np.linspace(0.0, 180.0, size, False)
    sinogram_circle = radon(image, theta, circle=True)

    def argmax_shape(a):
        return np.unravel_index(np.argmax(a), a.shape)

    print('\n\targmax of circle:', argmax_shape(sinogram_circle))
    sinogram_square = radon(image, theta, circle=False)
    print('\targmax of square:', argmax_shape(sinogram_square))
    sinogram_circle_to_square = _sinogram_circle_to_square(sinogram_circle)
    print('\targmax of circle to square:', argmax_shape(sinogram_circle_to_square))
    error = abs(sinogram_square - sinogram_circle_to_square)
    print(np.mean(error), np.max(error))
    assert argmax_shape(sinogram_square) == argmax_shape(sinogram_circle_to_square)


@pytest.mark.parametrize("size", (50, 51))
def test_sinogram_circle_to_square(size):
    check_sinogram_circle_to_square(size)


def check_radon_iradon_circle(interpolation, shape, output_size):
    # Forward and inverse radon on synthetic data
    image = _random_circle(shape)
    radius = min(shape) // 2
    sinogram_rectangle = radon(image, circle=False)
    reconstruction_rectangle = iradon(
        sinogram_rectangle,
        output_size=output_size,
        interpolation=interpolation,
        circle=False,
    )
    sinogram_circle = radon(image, circle=True)
    reconstruction_circle = iradon(
        sinogram_circle,
        output_size=output_size,
        interpolation=interpolation,
        circle=True,
    )
    # Crop rectangular reconstruction to match circle=True reconstruction
    width = reconstruction_circle.shape[0]
    excess = int(np.ceil((reconstruction_rectangle.shape[0] - width) / 2))
    s = np.s_[excess : width + excess, excess : width + excess]
    reconstruction_rectangle = reconstruction_rectangle[s]
    # Find the reconstruction circle, set reconstruction to zero outside
    c0, c1 = np.ogrid[0:width, 0:width]
    r = np.sqrt((c0 - width // 2) ** 2 + (c1 - width // 2) ** 2)
    reconstruction_rectangle[r > radius] = 0.0
    print(reconstruction_circle.shape)
    print(reconstruction_rectangle.shape)
    np.allclose(reconstruction_rectangle, reconstruction_circle)


# if adding more shapes to test data, you might want to look at commit d0f2bac3f
shapes_radon_iradon_circle = ((61, 79),)
interpolations = ('nearest', 'linear')
output_sizes = (
    None,
    min(shapes_radon_iradon_circle[0]),
    max(shapes_radon_iradon_circle[0]),
    97,
)


@pytest.mark.parametrize(
    "shape, interpolation, output_size",
    itertools.product(shapes_radon_iradon_circle, interpolations, output_sizes),
)
def test_radon_iradon_circle(shape, interpolation, output_size):
    check_radon_iradon_circle(interpolation, shape, output_size)


def test_order_angles_golden_ratio():
    from skimage.transform.radon_transform import order_angles_golden_ratio

    np.random.seed(1231)
    lengths = [1, 4, 10, 180]
    for l in lengths:
        theta_ordered = np.linspace(0, 180, l, endpoint=False)
        theta_random = np.random.uniform(0, 180, l)
        for theta in (theta_random, theta_ordered):
            indices = [x for x in order_angles_golden_ratio(theta)]
            # no duplicate indices allowed
            assert len(indices) == len(set(indices))


@run_in_parallel()
def test_iradon_sart():
    debug = False

    image = rescale(
        PHANTOM, 0.8, mode='reflect', channel_axis=None, anti_aliasing=False
    )
    theta_ordered = np.linspace(0.0, 180.0, image.shape[0], endpoint=False)
    theta_missing_wedge = np.linspace(0.0, 150.0, image.shape[0], endpoint=True)
    for theta, error_factor in ((theta_ordered, 1.0), (theta_missing_wedge, 2.0)):
        sinogram = radon(image, theta, circle=True)
        reconstructed = iradon_sart(sinogram, theta)

        if debug and has_mpl:
            from matplotlib import pyplot as plt

            plt.figure()
            plt.subplot(221)
            plt.imshow(image, interpolation='nearest')
            plt.subplot(222)
            plt.imshow(sinogram, interpolation='nearest')
            plt.subplot(223)
            plt.imshow(reconstructed, interpolation='nearest')
            plt.subplot(224)
            plt.imshow(reconstructed - image, interpolation='nearest')
            plt.show()

        delta = np.mean(np.abs(reconstructed - image))
        print('delta (1 iteration) =', delta)
        assert delta < 0.02 * error_factor
        reconstructed = iradon_sart(sinogram, theta, reconstructed)
        delta = np.mean(np.abs(reconstructed - image))
        print('delta (2 iterations) =', delta)
        assert delta < 0.014 * error_factor
        reconstructed = iradon_sart(sinogram, theta, clip=(0, 1))
        delta = np.mean(np.abs(reconstructed - image))
        print('delta (1 iteration, clip) =', delta)
        assert delta < 0.018 * error_factor

        np.random.seed(1239867)
        shifts = np.random.uniform(-3, 3, sinogram.shape[1])
        x = np.arange(sinogram.shape[0])
        sinogram_shifted = np.vstack(
            [
                np.interp(x + shifts[i], x, sinogram[:, i])
                for i in range(sinogram.shape[1])
            ]
        ).T
        reconstructed = iradon_sart(sinogram_shifted, theta, projection_shifts=shifts)
        if debug and has_mpl:
            from matplotlib import pyplot as plt

            plt.figure()
            plt.subplot(221)
            plt.imshow(image, interpolation='nearest')
            plt.subplot(222)
            plt.imshow(sinogram_shifted, interpolation='nearest')
            plt.subplot(223)
            plt.imshow(reconstructed, interpolation='nearest')
            plt.subplot(224)
            plt.imshow(reconstructed - image, interpolation='nearest')
            plt.show()

        delta = np.mean(np.abs(reconstructed - image))
        print('delta (1 iteration, shifted sinogram) =', delta)
        assert delta < 0.022 * error_factor


@pytest.mark.parametrize("preserve_range", [True, False])
def test_iradon_dtype(preserve_range):
    sinogram = np.zeros((16, 1), dtype=int)
    sinogram[8, 0] = 1.0
    sinogram64 = sinogram.astype('float64')
    sinogram32 = sinogram.astype('float32')

    assert iradon(sinogram, theta=[0], preserve_range=preserve_range).dtype == 'float64'
    assert (
        iradon(sinogram64, theta=[0], preserve_range=preserve_range).dtype
        == sinogram64.dtype
    )
    assert (
        iradon(sinogram32, theta=[0], preserve_range=preserve_range).dtype
        == sinogram32.dtype
    )


def test_radon_dtype():
    img = convert_to_float(PHANTOM, False)
    img32 = img.astype(np.float32)

    assert radon(img).dtype == img.dtype
    assert radon(img32).dtype == img32.dtype


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_iradon_sart_dtype(dtype):
    sinogram = np.zeros((16, 1), dtype=int)
    sinogram[8, 0] = 1.0
    sinogram64 = sinogram.astype('float64')
    sinogram32 = sinogram.astype('float32')

    with expected_warnings(['Input data is cast to float']):
        assert iradon_sart(sinogram, theta=[0]).dtype == 'float64'

    assert iradon_sart(sinogram64, theta=[0]).dtype == sinogram64.dtype
    assert iradon_sart(sinogram32, theta=[0]).dtype == sinogram32.dtype

    assert iradon_sart(sinogram, theta=[0], dtype=dtype).dtype == dtype
    assert iradon_sart(sinogram32, theta=[0], dtype=dtype).dtype == dtype
    assert iradon_sart(sinogram64, theta=[0], dtype=dtype).dtype == dtype


def test_iradon_sart_wrong_dtype():
    sinogram = np.zeros((16, 1))

    with pytest.raises(ValueError):
        iradon_sart(sinogram, dtype=int)


def test_iradon_rampfilter_bias_circular_phantom():
    """
    test that a uniform circular phantom has a small reconstruction bias using
    the ramp filter
    """
    pixels = 128
    xy = np.arange(-pixels / 2, pixels / 2) + 0.5
    x, y = np.meshgrid(xy, xy)
    image = x**2 + y**2 <= (pixels / 4) ** 2

    theta = np.linspace(0.0, 180.0, max(image.shape), endpoint=False)
    sinogram = radon(image, theta=theta)

    reconstruction_fbp = iradon(sinogram, theta=theta)
    error = reconstruction_fbp - image

    tol = 5e-5
    roi_err = np.abs(np.mean(error))
    assert roi_err < tol



from contextlib import contextmanager
import numpy as np
import pytest

from skimage._shared._dependency_checks import has_mpl
from skimage import io
from skimage.io import manage_plugins
from skimage._shared.testing import fetch, assert_stacklevel


priority_plugin = 'pil'


@pytest.fixture(autouse=True)
def _use_pil_plugin():
    io.use_plugin('pil')
    yield
    io.reset_plugins()


@contextmanager
def protect_preferred_plugins():
    """Contexts where `preferred_plugins` can be modified w/o side-effects."""
    preferred_plugins = manage_plugins.preferred_plugins.copy()
    try:
        yield
    finally:
        manage_plugins.preferred_plugins = preferred_plugins


def test_failed_use():
    with pytest.raises(ValueError):
        manage_plugins.use_plugin('asd')


@pytest.mark.skipif(not has_mpl, reason="matplotlib not installed")
def test_use_priority():
    manage_plugins.use_plugin(priority_plugin)
    plug, func = manage_plugins.plugin_store['imread'][0]
    np.testing.assert_equal(plug, priority_plugin)

    manage_plugins.use_plugin('matplotlib')
    plug, func = manage_plugins.plugin_store['imread'][0]
    np.testing.assert_equal(plug, 'matplotlib')


@pytest.mark.skipif(not has_mpl, reason="matplotlib not installed")
def test_load_preferred_plugins_all():
    from skimage.io._plugins import pil_plugin, matplotlib_plugin

    with protect_preferred_plugins():
        manage_plugins.preferred_plugins = {'all': ['pil'], 'imshow': ['matplotlib']}
        manage_plugins.reset_plugins()

        for plugin_type in ('imread', 'imsave'):
            plug, func = manage_plugins.plugin_store[plugin_type][0]
            assert func == getattr(pil_plugin, plugin_type)
        plug, func = manage_plugins.plugin_store['imshow'][0]
        assert func == getattr(matplotlib_plugin, 'imshow')


@pytest.mark.skipif(not has_mpl, reason="matplotlib not installed")
def test_load_preferred_plugins_imread():
    from skimage.io._plugins import pil_plugin, matplotlib_plugin

    with protect_preferred_plugins():
        manage_plugins.preferred_plugins['imread'] = ['pil']
        manage_plugins.reset_plugins()

        plug, func = manage_plugins.plugin_store['imread'][0]
        assert func == pil_plugin.imread
        plug, func = manage_plugins.plugin_store['imshow'][0]
        assert func == matplotlib_plugin.imshow, func.__module__


@pytest.mark.parametrize(
    ("func", "args"),
    [
        (io.use_plugin, ["imageio"]),
        (io.call_plugin, ["imread", fetch("data/camera.png")]),
        (io.plugin_info, ["imageio"]),
        (io.plugin_order, tuple()),
        (io.reset_plugins, tuple()),
        (io.find_available_plugins, tuple()),
        (getattr, [io, "available_plugins"]),
    ],
)
def test_deprecation_warnings_on_plugin_funcs(func, args):
    regex = ".*use `imageio` or other I/O packages directly.*"
    with pytest.warns(FutureWarning, match=regex) as record:
        func(*args)
    assert len(record) == 1
    assert_stacklevel(record, offset=-2)



import numpy as np
import pytest
from scipy import ndimage as ndi
from numpy.testing import assert_allclose, assert_array_equal, assert_equal

from skimage import color, data, transform
from skimage._shared._warnings import expected_warnings
from skimage._shared.testing import fetch, assert_stacklevel
from skimage.morphology import gray, footprints, footprint_rectangle
from skimage.util import img_as_uint, img_as_ubyte


@pytest.fixture
def cam_image():
    from skimage import data

    return np.ascontiguousarray(data.camera()[64:112, 64:96])


@pytest.fixture
def cell3d_image():
    from skimage import data

    return np.ascontiguousarray(data.cells3d()[30:48, 0, 20:36, 20:32])


gray_morphology_funcs = (
    gray.erosion,
    gray.dilation,
    gray.opening,
    gray.closing,
    gray.white_tophat,
    gray.black_tophat,
)


class TestMorphology:
    # These expected outputs were generated with skimage v0.22.0 + PR #6695
    # using:
    #
    #   from skimage.morphology.tests.test_gray import TestMorphology
    #   import numpy as np
    #   output = TestMorphology()._build_expected_output()
    #   np.savez_compressed('gray_morph_output.npz', **output)

    def _build_expected_output(self):
        def square(n):
            return footprint_rectangle((n, n))

        footprints_2D = (
            square,
            footprints.diamond,
            footprints.disk,
            footprints.star,
        )

        image = img_as_ubyte(
            transform.downscale_local_mean(color.rgb2gray(data.coffee()), (20, 20))
        )

        output = {}
        for n in range(1, 4):
            for strel in footprints_2D:
                for func in gray_morphology_funcs:
                    key = f'{strel.__name__}_{n}_{func.__name__}'
                    output[key] = func(image, strel(n))

        return output

    def test_gray_morphology(self):
        expected = dict(np.load(fetch('data/gray_morph_output.npz')))
        calculated = self._build_expected_output()
        assert_equal(expected, calculated)

    def test_gray_closing_extensive(self):
        img = data.coins()
        footprint = np.array([[0, 0, 1], [0, 1, 1], [1, 1, 1]])

        # Default mode="reflect" is not extensive for backwards-compatibility
        result_default = gray.closing(img, footprint=footprint)
        assert not np.all(result_default >= img)

        result = gray.closing(img, footprint=footprint, mode="ignore")
        assert np.all(result >= img)

    def test_gray_opening_anti_extensive(self):
        img = data.coins()
        footprint = np.array([[0, 0, 1], [0, 1, 1], [1, 1, 1]])

        # Default mode="reflect" is not extensive for backwards-compatibility
        result_default = gray.opening(img, footprint=footprint)
        assert not np.all(result_default <= img)

        result_ignore = gray.opening(img, footprint=footprint, mode="ignore")
        assert np.all(result_ignore <= img)

    @pytest.mark.parametrize("func", gray_morphology_funcs)
    @pytest.mark.parametrize("mode", gray._SUPPORTED_MODES)
    def test_supported_mode(self, func, mode):
        img = np.ones((10, 10))
        func(img, mode=mode)

    @pytest.mark.parametrize("func", gray_morphology_funcs)
    @pytest.mark.parametrize("mode", ["", "symmetric", 3, None])
    def test_unsupported_mode(self, func, mode):
        img = np.ones((10, 10))
        with pytest.raises(ValueError, match="unsupported mode"):
            func(img, mode=mode)


class TestEccentricStructuringElements:
    def setup_class(self):
        self.black_pixel = 255 * np.ones((6, 6), dtype=np.uint8)
        self.black_pixel[2, 2] = 0
        self.white_pixel = 255 - self.black_pixel
        self.footprints = [
            footprint_rectangle((2, 2)),
            footprint_rectangle((2, 1)),
            footprint_rectangle((2, 1)),
        ]

    def test_dilate_erode_symmetry(self):
        for s in self.footprints:
            c = gray.erosion(self.black_pixel, s)
            d = gray.dilation(self.white_pixel, s)
            assert np.all(c == (255 - d))

    def test_open_black_pixel(self):
        for s in self.footprints:
            gray_open = gray.opening(self.black_pixel, s)
            assert np.all(gray_open == self.black_pixel)

    def test_close_white_pixel(self):
        for s in self.footprints:
            gray_close = gray.closing(self.white_pixel, s)
            assert np.all(gray_close == self.white_pixel)

    def test_open_white_pixel(self):
        for s in self.footprints:
            assert np.all(gray.opening(self.white_pixel, s) == 0)

    def test_close_black_pixel(self):
        for s in self.footprints:
            assert np.all(gray.closing(self.black_pixel, s) == 255)

    def test_white_tophat_white_pixel(self):
        for s in self.footprints:
            tophat = gray.white_tophat(self.white_pixel, s)
            assert np.all(tophat == self.white_pixel)

    def test_black_tophat_black_pixel(self):
        for s in self.footprints:
            tophat = gray.black_tophat(self.black_pixel, s)
            assert np.all(tophat == self.white_pixel)

    def test_white_tophat_black_pixel(self):
        for s in self.footprints:
            tophat = gray.white_tophat(self.black_pixel, s)
            assert np.all(tophat == 0)

    def test_black_tophat_white_pixel(self):
        for s in self.footprints:
            tophat = gray.black_tophat(self.white_pixel, s)
            assert np.all(tophat == 0)


gray_functions = [
    gray.erosion,
    gray.dilation,
    gray.opening,
    gray.closing,
    gray.white_tophat,
    gray.black_tophat,
]


@pytest.mark.parametrize("function", gray_functions)
def test_default_footprint(function):
    strel = footprints.diamond(radius=1)
    image = np.array(
        [
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 1, 1, 1, 1, 1, 0, 0],
            [0, 0, 1, 1, 1, 1, 1, 1, 0, 0],
            [0, 0, 1, 1, 1, 1, 1, 1, 0, 0],
            [0, 0, 1, 1, 1, 0, 0, 1, 0, 0],
            [0, 0, 1, 1, 1, 0, 0, 1, 0, 0],
            [0, 0, 1, 1, 1, 0, 0, 1, 0, 0],
            [0, 0, 1, 1, 1, 1, 1, 1, 0, 0],
            [0, 0, 1, 1, 1, 1, 1, 1, 0, 0],
            [0, 0, 1, 1, 1, 1, 1, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ],
        np.uint8,
    )
    im_expected = function(image, strel)
    im_test = function(image)
    assert_array_equal(im_expected, im_test)


def test_3d_fallback_default_footprint():
    # 3x3x3 cube inside a 7x7x7 image:
    image = np.zeros((7, 7, 7), bool)
    image[2:-2, 2:-2, 2:-2] = 1

    opened = gray.opening(image)

    # expect a "hyper-cross" centered in the 5x5x5:
    image_expected = np.zeros((7, 7, 7), dtype=bool)
    image_expected[2:5, 2:5, 2:5] = ndi.generate_binary_structure(3, 1)
    assert_array_equal(opened, image_expected)


gray_3d_fallback_functions = [gray.closing, gray.opening]


@pytest.mark.parametrize("function", gray_3d_fallback_functions)
def test_3d_fallback_cube_footprint(function):
    # 3x3x3 cube inside a 7x7x7 image:
    image = np.zeros((7, 7, 7), bool)
    image[2:-2, 2:-2, 2:-2] = 1

    cube = np.ones((3, 3, 3), dtype=np.uint8)

    new_image = function(image, cube)
    assert_array_equal(new_image, image)


def test_3d_fallback_white_tophat():
    image = np.zeros((7, 7, 7), dtype=bool)
    image[2, 2:4, 2:4] = 1
    image[3, 2:5, 2:5] = 1
    image[4, 3:5, 3:5] = 1

    with expected_warnings([r'operator.*deprecated|\A\Z']):
        new_image = gray.white_tophat(image)
    footprint = ndi.generate_binary_structure(3, 1)
    with expected_warnings([r'operator.*deprecated|\A\Z']):
        image_expected = ndi.white_tophat(
            image.view(dtype=np.uint8), footprint=footprint
        )
    assert_array_equal(new_image, image_expected)


def test_3d_fallback_black_tophat():
    image = np.ones((7, 7, 7), dtype=bool)
    image[2, 2:4, 2:4] = 0
    image[3, 2:5, 2:5] = 0
    image[4, 3:5, 3:5] = 0

    with expected_warnings([r'operator.*deprecated|\A\Z']):
        new_image = gray.black_tophat(image)
    footprint = ndi.generate_binary_structure(3, 1)
    with expected_warnings([r'operator.*deprecated|\A\Z']):
        image_expected = ndi.black_tophat(
            image.view(dtype=np.uint8), footprint=footprint
        )
    assert_array_equal(new_image, image_expected)


def test_2d_ndimage_equivalence():
    image = np.zeros((9, 9), np.uint8)
    image[2:-2, 2:-2] = 128
    image[3:-3, 3:-3] = 196
    image[4, 4] = 255

    opened = gray.opening(image)
    closed = gray.closing(image)

    footprint = ndi.generate_binary_structure(2, 1)
    ndimage_opened = ndi.grey_opening(image, footprint=footprint)
    ndimage_closed = ndi.grey_closing(image, footprint=footprint)

    assert_array_equal(opened, ndimage_opened)
    assert_array_equal(closed, ndimage_closed)


# float test images
im = np.array(
    [
        [0.55, 0.72, 0.6, 0.54, 0.42],
        [0.65, 0.44, 0.89, 0.96, 0.38],
        [0.79, 0.53, 0.57, 0.93, 0.07],
        [0.09, 0.02, 0.83, 0.78, 0.87],
        [0.98, 0.8, 0.46, 0.78, 0.12],
    ]
)

eroded = np.array(
    [
        [0.55, 0.44, 0.54, 0.42, 0.38],
        [0.44, 0.44, 0.44, 0.38, 0.07],
        [0.09, 0.02, 0.53, 0.07, 0.07],
        [0.02, 0.02, 0.02, 0.78, 0.07],
        [0.09, 0.02, 0.46, 0.12, 0.12],
    ]
)

dilated = np.array(
    [
        [0.72, 0.72, 0.89, 0.96, 0.54],
        [0.79, 0.89, 0.96, 0.96, 0.96],
        [0.79, 0.79, 0.93, 0.96, 0.93],
        [0.98, 0.83, 0.83, 0.93, 0.87],
        [0.98, 0.98, 0.83, 0.78, 0.87],
    ]
)

opened = np.array(
    [
        [0.55, 0.55, 0.54, 0.54, 0.42],
        [0.55, 0.44, 0.54, 0.44, 0.38],
        [0.44, 0.53, 0.53, 0.78, 0.07],
        [0.09, 0.02, 0.78, 0.78, 0.78],
        [0.09, 0.46, 0.46, 0.78, 0.12],
    ]
)

closed = np.array(
    [
        [0.72, 0.72, 0.72, 0.54, 0.54],
        [0.72, 0.72, 0.89, 0.96, 0.54],
        [0.79, 0.79, 0.79, 0.93, 0.87],
        [0.79, 0.79, 0.83, 0.78, 0.87],
        [0.98, 0.83, 0.78, 0.78, 0.78],
    ]
)


def test_float():
    assert_allclose(gray.erosion(im), eroded)
    assert_allclose(gray.dilation(im), dilated)
    assert_allclose(gray.opening(im), opened)
    assert_allclose(gray.closing(im), closed)


def test_uint16():
    im16, eroded16, dilated16, opened16, closed16 = map(
        img_as_uint, [im, eroded, dilated, opened, closed]
    )
    assert_allclose(gray.erosion(im16), eroded16)
    assert_allclose(gray.dilation(im16), dilated16)
    assert_allclose(gray.opening(im16), opened16)
    assert_allclose(gray.closing(im16), closed16)


def test_discontiguous_out_array():
    image = np.array([[5, 6, 2], [7, 2, 2], [3, 5, 1]], np.uint8)
    out_array_big = np.zeros((5, 5), np.uint8)
    out_array = out_array_big[::2, ::2]
    expected_dilation = np.array(
        [
            [7, 0, 6, 0, 6],
            [0, 0, 0, 0, 0],
            [7, 0, 7, 0, 2],
            [0, 0, 0, 0, 0],
            [7, 0, 5, 0, 5],
        ],
        np.uint8,
    )
    expected_erosion = np.array(
        [
            [5, 0, 2, 0, 2],
            [0, 0, 0, 0, 0],
            [2, 0, 2, 0, 1],
            [0, 0, 0, 0, 0],
            [3, 0, 1, 0, 1],
        ],
        np.uint8,
    )
    gray.dilation(image, out=out_array)
    assert_array_equal(out_array_big, expected_dilation)
    gray.erosion(image, out=out_array)
    assert_array_equal(out_array_big, expected_erosion)


def test_1d_erosion():
    image = np.array([1, 2, 3, 2, 1])
    expected = np.array([1, 1, 2, 1, 1])
    eroded = gray.erosion(image)
    assert_array_equal(eroded, expected)


@pytest.mark.parametrize(
    "function",
    ["erosion", "dilation", "closing", "opening", "white_tophat", "black_tophat"],
)
@pytest.mark.parametrize("nrows", [3, 7, 11])
@pytest.mark.parametrize("ncols", [3, 7, 11])
@pytest.mark.parametrize("decomposition", ['separable', 'sequence'])
def test_rectangle_decomposition(cam_image, function, nrows, ncols, decomposition):
    """Validate footprint decomposition for various shapes.

    comparison is made to the case without decomposition.
    """
    footprint_ndarray = footprint_rectangle((nrows, ncols), decomposition=None)
    footprint = footprint_rectangle((nrows, ncols), decomposition=decomposition)
    func = getattr(gray, function)
    expected = func(cam_image, footprint=footprint_ndarray)
    out = func(cam_image, footprint=footprint)
    assert_array_equal(expected, out)


@pytest.mark.parametrize(
    "function",
    ["erosion", "dilation", "closing", "opening", "white_tophat", "black_tophat"],
)
@pytest.mark.parametrize("radius", (2, 3))
@pytest.mark.parametrize("decomposition", ['sequence'])
def test_diamond_decomposition(cam_image, function, radius, decomposition):
    """Validate footprint decomposition for various shapes.

    comparison is made to the case without decomposition.
    """
    footprint_ndarray = footprints.diamond(radius, decomposition=None)
    footprint = footprints.diamond(radius, decomposition=decomposition)
    func = getattr(gray, function)
    expected = func(cam_image, footprint=footprint_ndarray)
    out = func(cam_image, footprint=footprint)
    assert_array_equal(expected, out)


@pytest.mark.parametrize(
    "function",
    ["erosion", "dilation", "closing", "opening", "white_tophat", "black_tophat"],
)
@pytest.mark.parametrize("m", (0, 1, 3, 5))
@pytest.mark.parametrize("n", (0, 1, 2, 3))
@pytest.mark.parametrize("decomposition", ['sequence'])
@pytest.mark.filterwarnings(
    "ignore:.*falling back to decomposition='separable':UserWarning:skimage"
)
def test_octagon_decomposition(cam_image, function, m, n, decomposition):
    """Validate footprint decomposition for various shapes.

    comparison is made to the case without decomposition.
    """
    if m == 0 and n == 0:
        with pytest.raises(ValueError):
            footprints.octagon(m, n, decomposition=decomposition)
    else:
        footprint_ndarray = footprints.octagon(m, n, decomposition=None)
        footprint = footprints.octagon(m, n, decomposition=decomposition)
        func = getattr(gray, function)
        expected = func(cam_image, footprint=footprint_ndarray)
        out = func(cam_image, footprint=footprint)
        assert_array_equal(expected, out)


@pytest.mark.parametrize(
    "function",
    ["erosion", "dilation", "closing", "opening", "white_tophat", "black_tophat"],
)
@pytest.mark.parametrize("shape", [(5, 5, 5), (5, 5, 7)])
@pytest.mark.parametrize("decomposition", ['separable', 'sequence'])
def test_cube_decomposition(cell3d_image, function, shape, decomposition):
    """Validate footprint decomposition for various shapes.

    comparison is made to the case without decomposition.
    """
    footprint_ndarray = footprint_rectangle(shape, decomposition=None)
    footprint = footprint_rectangle(shape, decomposition=decomposition)
    func = getattr(gray, function)
    expected = func(cell3d_image, footprint=footprint_ndarray)
    out = func(cell3d_image, footprint=footprint)
    assert_array_equal(expected, out)


@pytest.mark.parametrize(
    "function",
    ["erosion", "dilation", "closing", "opening", "white_tophat", "black_tophat"],
)
@pytest.mark.parametrize("radius", (3,))
@pytest.mark.parametrize("decomposition", ['sequence'])
def test_octahedron_decomposition(cell3d_image, function, radius, decomposition):
    """Validate footprint decomposition for various shapes.

    comparison is made to the case without decomposition.
    """
    footprint_ndarray = footprints.octahedron(radius, decomposition=None)
    footprint = footprints.octahedron(radius, decomposition=decomposition)
    func = getattr(gray, function)
    expected = func(cell3d_image, footprint=footprint_ndarray)
    out = func(cell3d_image, footprint=footprint)
    assert_array_equal(expected, out)


@pytest.mark.parametrize("func", [gray.erosion, gray.dilation])
@pytest.mark.parametrize("name", ["shift_x", "shift_y"])
@pytest.mark.parametrize("value", [True, False, None])
def test_deprecated_shift(func, name, value):
    img = np.ones(10)
    func(img)  # Shouldn't warn

    regex = "`shift_x` and `shift_y` are deprecated"
    with pytest.warns(FutureWarning, match=regex) as record:
        func(img, **{name: value})
    assert_stacklevel(record)



import numpy as np

from skimage._shared import testing
from skimage._shared.testing import assert_equal
from skimage.util.shape import view_as_blocks, view_as_windows


def test_view_as_blocks_block_not_a_tuple():
    A = np.arange(10)
    with testing.raises(TypeError):
        view_as_blocks(A, [5])


def test_view_as_blocks_negative_shape():
    A = np.arange(10)
    with testing.raises(ValueError):
        view_as_blocks(A, (-2,))


def test_view_as_blocks_block_too_large():
    A = np.arange(10)
    with testing.raises(ValueError):
        view_as_blocks(A, (11,))


def test_view_as_blocks_wrong_block_dimension():
    A = np.arange(10)
    with testing.raises(ValueError):
        view_as_blocks(A, (2, 2))


def test_view_as_blocks_1D_array_wrong_block_shape():
    A = np.arange(10)
    with testing.raises(ValueError):
        view_as_blocks(A, (3,))


def test_view_as_blocks_1D_array():
    A = np.arange(10)
    B = view_as_blocks(A, (5,))
    assert_equal(B, np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]))


def test_view_as_blocks_2D_array():
    A = np.arange(4 * 4).reshape(4, 4)
    B = view_as_blocks(A, (2, 2))
    assert_equal(B[0, 1], np.array([[2, 3], [6, 7]]))
    assert_equal(B[1, 0, 1, 1], 13)


def test_view_as_blocks_3D_array():
    A = np.arange(4 * 4 * 6).reshape(4, 4, 6)
    B = view_as_blocks(A, (1, 2, 2))
    assert_equal(B.shape, (4, 2, 3, 1, 2, 2))
    assert_equal(
        B[2:, 0, 2], np.array([[[[52, 53], [58, 59]]], [[[76, 77], [82, 83]]]])
    )


def test_view_as_windows_input_not_array():
    A = [1, 2, 3, 4, 5]
    with testing.raises(TypeError):
        view_as_windows(A, (2,))


def test_view_as_windows_wrong_window_dimension():
    A = np.arange(10)
    with testing.raises(ValueError):
        view_as_windows(A, (2, 2))


def test_view_as_windows_negative_window_length():
    A = np.arange(10)
    with testing.raises(ValueError):
        view_as_windows(A, (-1,))


def test_view_as_windows_window_too_large():
    A = np.arange(10)
    with testing.raises(ValueError):
        view_as_windows(A, (11,))


def test_view_as_windows_step_below_one():
    A = np.arange(10)
    with testing.raises(ValueError):
        view_as_windows(A, (11,), step=0.9)


def test_view_as_windows_1D():
    A = np.arange(10)
    window_shape = (3,)
    B = view_as_windows(A, window_shape)
    assert_equal(
        B,
        np.array(
            [
                [0, 1, 2],
                [1, 2, 3],
                [2, 3, 4],
                [3, 4, 5],
                [4, 5, 6],
                [5, 6, 7],
                [6, 7, 8],
                [7, 8, 9],
            ]
        ),
    )


def test_view_as_windows_2D():
    A = np.arange(5 * 4).reshape(5, 4)
    window_shape = (4, 3)
    B = view_as_windows(A, window_shape)
    assert_equal(B.shape, (2, 2, 4, 3))
    assert_equal(
        B,
        np.array(
            [
                [
                    [[0, 1, 2], [4, 5, 6], [8, 9, 10], [12, 13, 14]],
                    [[1, 2, 3], [5, 6, 7], [9, 10, 11], [13, 14, 15]],
                ],
                [
                    [[4, 5, 6], [8, 9, 10], [12, 13, 14], [16, 17, 18]],
                    [[5, 6, 7], [9, 10, 11], [13, 14, 15], [17, 18, 19]],
                ],
            ]
        ),
    )


def test_view_as_windows_with_skip():
    A = np.arange(20).reshape((5, 4))
    B = view_as_windows(A, 2, step=2)
    assert_equal(
        B,
        [
            [[[0, 1], [4, 5]], [[2, 3], [6, 7]]],
            [[[8, 9], [12, 13]], [[10, 11], [14, 15]]],
        ],
    )

    C = view_as_windows(A, 2, step=4)
    assert_equal(C.shape, (1, 1, 2, 2))


def test_views_non_contiguous():
    A = np.arange(16).reshape((4, 4))
    A = A[::2, :]

    res_b = view_as_blocks(A, (2, 2))
    res_w = view_as_windows(A, (2, 2))
    print(res_b)
    print(res_w)

    expected_b = [[[[0, 1], [8, 9]], [[2, 3], [10, 11]]]]

    expected_w = [[[[0, 1], [8, 9]], [[1, 2], [9, 10]], [[2, 3], [10, 11]]]]

    assert_equal(res_b, expected_b)
    assert_equal(res_w, expected_w)


def test_view_as_windows_step_tuple():
    A = np.arange(24).reshape((6, 4))
    B = view_as_windows(A, (3, 2), step=3)
    assert B.shape == (2, 1, 3, 2)
    assert B.size != A.size

    C = view_as_windows(A, (3, 2), step=(3, 2))
    assert C.shape == (2, 2, 3, 2)
    assert C.size == A.size

    assert_equal(
        C,
        [
            [[[0, 1], [4, 5], [8, 9]], [[2, 3], [6, 7], [10, 11]]],
            [[[12, 13], [16, 17], [20, 21]], [[14, 15], [18, 19], [22, 23]]],
        ],
    )



import numpy as np
import scipy as sp
import pytest

from skimage.metrics import (
    adapted_rand_error,
    variation_of_information,
    contingency_table,
)

from skimage._shared.testing import (
    assert_equal,
    assert_almost_equal,
    assert_array_equal,
)


@pytest.mark.parametrize("sparse_type", ["matrix", "array"])
def test_contingency_table(sparse_type):
    im_true = np.array([1, 2, 3, 4])
    im_test = np.array([1, 1, 8, 8])

    table1 = np.array(
        [
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.25, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.25, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.25],
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.25],
        ]
    )

    sparse_table2 = contingency_table(
        im_true, im_test, normalize=True, sparse_type=sparse_type
    )
    table2 = sparse_table2.toarray()
    assert_array_equal(table1, table2)


def test_contingency_table_sparse_type():
    im_true = np.array([1, 2, 3, 4])
    im_test = np.array([1, 1, 8, 8])

    result = contingency_table(im_true, im_test)
    assert isinstance(result, sp.sparse.csr_matrix)

    result = contingency_table(im_true, im_test, sparse_type="matrix")
    assert isinstance(result, sp.sparse.csr_matrix)

    result = contingency_table(im_true, im_test, sparse_type="array")
    assert isinstance(result, sp.sparse.csr_array)

    with pytest.raises(ValueError, match="`sparse_type` must be 'array' or 'matrix'"):
        contingency_table(im_true, im_test, sparse_type="unknown")


def test_vi():
    im_true = np.array([1, 2, 3, 4])
    im_test = np.array([1, 1, 8, 8])
    assert_equal(np.sum(variation_of_information(im_true, im_test)), 1)


def test_vi_ignore_labels():
    im1 = np.array([[1, 0], [2, 3]], dtype='uint8')
    im2 = np.array([[1, 1], [1, 0]], dtype='uint8')

    false_splits, false_merges = variation_of_information(im1, im2, ignore_labels=[0])
    assert (false_splits, false_merges) == (0, 2 / 3)


def test_are():
    im_true = np.array([[2, 1], [1, 2]])
    im_test = np.array([[1, 2], [3, 1]])
    assert_almost_equal(adapted_rand_error(im_true, im_test), (0.3333333, 0.5, 1.0))
    assert_almost_equal(adapted_rand_error(im_true, im_test, alpha=0), (0, 0.5, 1.0))
    assert_almost_equal(adapted_rand_error(im_true, im_test, alpha=1), (0.5, 0.5, 1.0))

    with pytest.raises(ValueError):
        adapted_rand_error(im_true, im_test, alpha=1.01)
    with pytest.raises(ValueError):
        adapted_rand_error(im_true, im_test, alpha=-0.01)



import skimage.data as data
from skimage.feature import Cascade


def test_detector_astronaut():
    # Load the trained file from the module root.
    trained_file = data.lbp_frontal_face_cascade_filename()

    # Initialize the detector cascade.
    detector = Cascade(trained_file)

    img = data.astronaut()

    detected = detector.detect_multi_scale(
        img=img, scale_factor=1.2, step_ratio=1, min_size=(60, 60), max_size=(123, 123)
    )

    assert len(detected) == 1, 'One face should be detected.'



import numpy as np
import pytest

import skimage as ski
from skimage.transform import ThinPlateSplineTransform

SRC = np.array([[0, 0], [0, 5], [5, 5], [5, 0]])

DST = np.array([[5, 0], [0, 0], [0, 5], [5, 5]])


class TestThinPlateSplineTransform:
    def test_call_before_estimation(self):
        tps = ThinPlateSplineTransform()
        assert tps.src is None
        with pytest.raises(ValueError, match="Transformation is undefined"):
            tps(SRC)

    def test_call_invalid_coords_shape(self):
        tps = ThinPlateSplineTransform()
        tps.estimate(SRC, DST)
        coords = np.array([1, 2, 3])
        with pytest.raises(
            ValueError, match=r"Input `coords` must have shape \(N, 2\)"
        ):
            tps(coords)

    def test_call_on_SRC(self):
        tps = ThinPlateSplineTransform()
        tps.estimate(SRC, DST)
        result = tps(SRC)
        np.testing.assert_allclose(result, DST, atol=1e-15)

    def test_tps_transform_inverse(self):
        tps = ThinPlateSplineTransform()
        tps.estimate(SRC, DST)
        with pytest.raises(NotImplementedError):
            tps.inverse()

    def test_tps_estimation_faulty_input(self):
        src = np.array([[0, 0], [0, 5], [5, 5], [5, 0]])
        dst = np.array([[5, 0], [0, 0], [0, 5]])

        tps = ThinPlateSplineTransform()
        assert tps.src is None

        with pytest.raises(ValueError, match="Shape of `src` and `dst` didn't match"):
            tps.estimate(src, dst)

        less_than_3pts = np.array([[0, 0], [0, 5]])
        with pytest.raises(ValueError, match="Need at least 3 points"):
            tps.estimate(less_than_3pts, dst)
        with pytest.raises(ValueError, match="Need at least 3 points"):
            tps.estimate(src, less_than_3pts)
        with pytest.raises(ValueError, match="Need at least 3 points"):
            tps.estimate(less_than_3pts, less_than_3pts)

        not_2d = np.array([0, 1, 2, 3])
        with pytest.raises(ValueError, match=".*`src` must be a 2-dimensional array"):
            tps.estimate(not_2d, dst)
        with pytest.raises(ValueError, match=".*`dst` must be a 2-dimensional array"):
            tps.estimate(src, not_2d)

        # When the estimation fails, the instance attributes remain unchanged
        assert tps.src is None

    def test_rotate(self):
        image = ski.data.astronaut()
        desired = ski.transform.rotate(image, angle=90)

        src = np.array([[0, 0], [0, 511], [511, 511], [511, 0]])
        dst = np.array([[511, 0], [0, 0], [0, 511], [511, 511]])
        tps = ThinPlateSplineTransform()
        tps.estimate(src, dst)
        result = ski.transform.warp(image, tps)

        np.testing.assert_allclose(result, desired, atol=1e-13)



"""Testing decorators module"""

import inspect
import re
import warnings

import pytest
from numpy.testing import assert_equal
from skimage._shared.testing import (
    doctest_skip_parser,
    run_in_parallel,
    assert_stacklevel,
)
from skimage._shared import testing
from skimage._shared._dependency_checks import is_wasm

from skimage._shared._warnings import expected_warnings
from warnings import warn


def test_skipper():
    def f():
        pass

    class c:
        def __init__(self):
            self.me = "I think, therefore..."

    docstring = """ Header

            >>> something # skip if not HAVE_AMODULE
            >>> something + else
            >>> a = 1 # skip if not HAVE_BMODULE
            >>> something2   # skip if HAVE_AMODULE
        """
    f.__doc__ = docstring
    c.__doc__ = docstring

    global HAVE_AMODULE, HAVE_BMODULE
    HAVE_AMODULE = False
    HAVE_BMODULE = True

    f2 = doctest_skip_parser(f)
    c2 = doctest_skip_parser(c)
    assert f is f2
    assert c is c2

    expected = """ Header

            >>> something # doctest: +SKIP
            >>> something + else
            >>> a = 1
            >>> something2
        """
    assert_equal(f2.__doc__, expected)
    assert_equal(c2.__doc__, expected)

    HAVE_AMODULE = True
    HAVE_BMODULE = False
    f.__doc__ = docstring
    c.__doc__ = docstring
    f2 = doctest_skip_parser(f)
    c2 = doctest_skip_parser(c)

    assert f is f2
    expected = """ Header

            >>> something
            >>> something + else
            >>> a = 1 # doctest: +SKIP
            >>> something2   # doctest: +SKIP
        """
    assert_equal(f2.__doc__, expected)
    assert_equal(c2.__doc__, expected)

    del HAVE_AMODULE
    f.__doc__ = docstring
    c.__doc__ = docstring
    with testing.raises(NameError):
        doctest_skip_parser(f)
    with testing.raises(NameError):
        doctest_skip_parser(c)


@pytest.mark.skipif(is_wasm, reason="Cannot start threads in WASM")
def test_run_in_parallel():
    state = []

    @run_in_parallel()
    def change_state1():
        state.append(None)

    change_state1()
    assert len(state) == 2

    @run_in_parallel(num_threads=1)
    def change_state2():
        state.append(None)

    change_state2()
    assert len(state) == 3

    @run_in_parallel(num_threads=3)
    def change_state3():
        state.append(None)

    change_state3()
    assert len(state) == 6


@pytest.mark.skipif(is_wasm, reason="Cannot run parallel code in WASM")
def test_parallel_warning():
    @run_in_parallel()
    def change_state_warns_fails():
        warn("Test warning for test parallel", stacklevel=2)

    with expected_warnings(['Test warning for test parallel']):
        change_state_warns_fails()

    @run_in_parallel(warnings_matching=['Test warning for test parallel'])
    def change_state_warns_passes():
        warn("Test warning for test parallel", stacklevel=2)

    change_state_warns_passes()


def test_expected_warnings_noop():
    # This will ensure the line beolow it behaves like a no-op
    with expected_warnings(['Expected warnings test']):
        # This should behave as a no-op
        with expected_warnings(None):
            warn('Expected warnings test')


class Test_assert_stacklevel:
    def raise_warning(self, *args, **kwargs):
        warnings.warn(*args, **kwargs)

    def test_correct_stacklevel(self):
        # Should pass if stacklevel is set correctly
        with pytest.warns(UserWarning, match="passes") as record:
            self.raise_warning("passes", UserWarning, stacklevel=2)
        assert_stacklevel(record)

    @pytest.mark.parametrize("level", [1, 3])
    def test_wrong_stacklevel(self, level):
        # AssertionError should be raised for wrong stacklevel
        with pytest.warns(UserWarning, match="wrong") as record:
            self.raise_warning("wrong", UserWarning, stacklevel=level)
        # Check that message contains expected line on right side
        line_number = inspect.currentframe().f_lineno - 2
        regex = ".*" + re.escape(f"Expected: {__file__}:{line_number}")
        with pytest.raises(AssertionError, match=regex):
            assert_stacklevel(record, offset=-5)



import pytest

# List of files that pytest should ignore
collect_ignore = [
    "io/_plugins",
]


@pytest.fixture(autouse=True)
def handle_np2():
    # TODO: remove when we require numpy >= 2
    try:
        import numpy as np

        np.set_printoptions(legacy="1.21")
    except ImportError:
        pass



import numpy as np
import pytest

from skimage._shared._dependency_checks import has_mpl
from skimage.feature.util import (
    FeatureDetector,
    DescriptorExtractor,
    _prepare_grayscale_input_2D,
    _mask_border_keypoints,
    plot_matched_features,
)


def test_feature_detector():
    with pytest.raises(NotImplementedError):
        FeatureDetector().detect(None)


def test_descriptor_extractor():
    with pytest.raises(NotImplementedError):
        DescriptorExtractor().extract(None, None)


def test_prepare_grayscale_input_2D():
    with pytest.raises(ValueError):
        _prepare_grayscale_input_2D(np.zeros((3, 3, 3)))
    with pytest.raises(ValueError):
        _prepare_grayscale_input_2D(np.zeros((3, 1)))
    with pytest.raises(ValueError):
        _prepare_grayscale_input_2D(np.zeros((3, 1, 1)))
    _prepare_grayscale_input_2D(np.zeros((3, 3)))
    _prepare_grayscale_input_2D(np.zeros((3, 3, 1)))
    _prepare_grayscale_input_2D(np.zeros((1, 3, 3)))


def test_mask_border_keypoints():
    keypoints = np.array([[0, 0], [1, 1], [2, 2], [3, 3], [4, 4]])
    np.testing.assert_equal(
        _mask_border_keypoints((10, 10), keypoints, 0), [1, 1, 1, 1, 1]
    )
    np.testing.assert_equal(
        _mask_border_keypoints((10, 10), keypoints, 2), [0, 0, 1, 1, 1]
    )
    np.testing.assert_equal(
        _mask_border_keypoints((4, 4), keypoints, 2), [0, 0, 1, 0, 0]
    )
    np.testing.assert_equal(
        _mask_border_keypoints((10, 10), keypoints, 5), [0, 0, 0, 0, 0]
    )
    np.testing.assert_equal(
        _mask_border_keypoints((10, 10), keypoints, 4), [0, 0, 0, 0, 1]
    )


@pytest.mark.skipif(not has_mpl, reason="Matplotlib not installed")
@pytest.mark.parametrize(
    "shapes",
    [
        ((10, 10), (10, 10)),
        ((10, 10), (12, 10)),
        ((10, 10), (10, 12)),
        ((10, 10), (12, 12)),
        ((12, 10), (10, 10)),
        ((10, 12), (10, 10)),
        ((12, 12), (10, 10)),
    ],
)
def test_plot_matched_features(shapes):
    from matplotlib import pyplot as plt
    from matplotlib import use

    use('Agg')

    fig, ax = plt.subplots()

    rng = np.random.default_rng(202410101501)
    keypoints0 = 10 * rng.random((10, 2))
    keypoints1 = 10 * rng.random((10, 2))
    idxs0 = rng.integers(10, size=10)
    idxs1 = rng.integers(10, size=10)
    matches = np.column_stack((idxs0, idxs1))

    shape0, shape1 = shapes
    img0 = np.zeros(shape0)
    img1 = np.zeros(shape1)
    plot_matched_features(
        img0,
        img1,
        keypoints0=keypoints0,
        keypoints1=keypoints1,
        matches=matches,
        ax=ax,
    )
    plot_matched_features(
        img0,
        img1,
        ax=ax,
        keypoints0=keypoints0,
        keypoints1=keypoints1,
        matches=matches,
        only_matches=True,
    )
    plot_matched_features(
        img0,
        img1,
        ax=ax,
        keypoints0=keypoints0,
        keypoints1=keypoints1,
        matches=matches,
        keypoints_color='r',
    )
    plot_matched_features(
        img0,
        img1,
        ax=ax,
        keypoints0=keypoints0,
        keypoints1=keypoints1,
        matches=matches,
        matches_color='r',
    )
    # Pass colors as random list of color strings
    rng = np.random.default_rng(202409281822)
    random_matches_color = [
        rng.choice(['C0', '#abc', 'aquamarine']) for _ in range(len(matches))
    ]
    plot_matched_features(
        img0,
        img1,
        ax=ax,
        keypoints0=keypoints0,
        keypoints1=keypoints1,
        matches=matches,
        matches_color=random_matches_color,
    )
    # Pass colors as single array of shape (len(matches), 3)
    plot_matched_features(
        img0,
        img1,
        ax=ax,
        keypoints0=keypoints0,
        keypoints1=keypoints1,
        matches=matches,
        matches_color=np.linspace((0, 0, 0), (1, 1, 1), num=len(matches)),
    )
    plot_matched_features(
        img0,
        img1,
        ax=ax,
        keypoints0=keypoints0,
        keypoints1=keypoints1,
        matches=matches,
        alignment='vertical',
    )
    plt.close()


@pytest.mark.skipif(not has_mpl, reason="Matplotlib not installed")
@pytest.mark.parametrize("matches_color", ([], ["C0"], ["C0", "C1"], np.arange(30)))
def test_plot_matched_features_color_error(matches_color):
    from matplotlib import pyplot as plt
    from matplotlib import use

    use('Agg')

    _, ax = plt.subplots()

    keypoints0 = 10 * np.random.rand(10, 2)
    keypoints1 = 10 * np.random.rand(10, 2)
    idxs0 = np.random.randint(10, size=10)
    idxs1 = np.random.randint(10, size=10)
    matches = np.column_stack((idxs0, idxs1))
    assert len(matches_color) != len(matches)

    img0 = np.zeros((10, 10))
    img1 = np.zeros_like(img0)

    regex = (
        '`matches_color` needs to be a single color '
        'or a sequence of length equal to the number of matches'
    )
    with pytest.raises(ValueError, match=regex):
        plot_matched_features(
            img0,
            img1,
            ax=ax,
            keypoints0=keypoints0,
            keypoints1=keypoints1,
            matches=matches,
            matches_color=matches_color,
        )


@pytest.mark.skipif(not has_mpl, reason="Matplotlib not installed")
def test_plot_matched_features_matplotlib_color_error():
    # Error is raised from matplotlib itself if we pass a sequence of correct length
    # but with values that aren't colors

    from matplotlib import pyplot as plt
    from matplotlib import use

    use('Agg')

    _, ax = plt.subplots()

    keypoints0 = 10 * np.random.rand(10, 2)
    keypoints1 = 10 * np.random.rand(10, 2)
    idxs0 = np.random.randint(10, size=10)
    idxs1 = np.random.randint(10, size=10)
    matches = np.column_stack((idxs0, idxs1))

    img0 = np.zeros((10, 10))
    img1 = np.zeros_like(img0)

    with pytest.raises(ValueError, match=".* not a valid value for color"):
        plot_matched_features(
            img0,
            img1,
            ax=ax,
            keypoints0=keypoints0,
            keypoints1=keypoints1,
            matches=matches,
            matches_color=np.arange(len(matches)),
        )



import numpy as np
import skimage.graph.mcp as mcp

from skimage._shared.testing import assert_array_equal, assert_almost_equal, parametrize
from skimage._shared._warnings import expected_warnings


np.random.seed(0)
a = np.ones((8, 8), dtype=np.float32)
a[1:-1, 1] = 0
a[1, 1:-1] = 0

warning_optional = r'|\A\Z'


def test_basic():
    with expected_warnings(['Upgrading NumPy' + warning_optional]):
        m = mcp.MCP(a, fully_connected=True)
    costs, traceback = m.find_costs([(1, 6)])
    return_path = m.traceback((7, 2))
    assert_array_equal(
        costs,
        [
            [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
            [1.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            [1.0, 0.0, 1.0, 2.0, 2.0, 2.0, 2.0, 2.0],
            [1.0, 0.0, 1.0, 2.0, 3.0, 3.0, 3.0, 3.0],
            [1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 4.0, 4.0],
            [1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 5.0],
            [1.0, 1.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
        ],
    )

    assert_array_equal(
        return_path,
        [
            (1, 6),
            (1, 5),
            (1, 4),
            (1, 3),
            (1, 2),
            (2, 1),
            (3, 1),
            (4, 1),
            (5, 1),
            (6, 1),
            (7, 2),
        ],
    )


def test_neg_inf():
    expected_costs = np.where(a == 1, np.inf, 0)
    expected_path = [
        (1, 6),
        (1, 5),
        (1, 4),
        (1, 3),
        (1, 2),
        (2, 1),
        (3, 1),
        (4, 1),
        (5, 1),
        (6, 1),
    ]
    test_neg = np.where(a == 1, -1, 0)
    test_inf = np.where(a == 1, np.inf, 0)
    with expected_warnings(['Upgrading NumPy' + warning_optional]):
        m = mcp.MCP(test_neg, fully_connected=True)
    costs, traceback = m.find_costs([(1, 6)])
    return_path = m.traceback((6, 1))
    assert_array_equal(costs, expected_costs)
    assert_array_equal(return_path, expected_path)
    with expected_warnings(['Upgrading NumPy' + warning_optional]):
        m = mcp.MCP(test_inf, fully_connected=True)
    costs, traceback = m.find_costs([(1, 6)])
    return_path = m.traceback((6, 1))
    assert_array_equal(costs, expected_costs)
    assert_array_equal(return_path, expected_path)


def test_route():
    with expected_warnings(['Upgrading NumPy' + warning_optional]):
        return_path, cost = mcp.route_through_array(a, (1, 6), (7, 2), geometric=True)
    assert_almost_equal(cost, np.sqrt(2) / 2)
    assert_array_equal(
        return_path,
        [
            (1, 6),
            (1, 5),
            (1, 4),
            (1, 3),
            (1, 2),
            (2, 1),
            (3, 1),
            (4, 1),
            (5, 1),
            (6, 1),
            (7, 2),
        ],
    )


def test_no_diagonal():
    with expected_warnings(['Upgrading NumPy' + warning_optional]):
        m = mcp.MCP(a, fully_connected=False)
    costs, traceback = m.find_costs([(1, 6)])
    return_path = m.traceback((7, 2))
    assert_array_equal(
        costs,
        [
            [2.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 2.0],
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
            [1.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 2.0],
            [1.0, 0.0, 1.0, 2.0, 2.0, 2.0, 2.0, 3.0],
            [1.0, 0.0, 1.0, 2.0, 3.0, 3.0, 3.0, 4.0],
            [1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 4.0, 5.0],
            [1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0],
            [2.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0],
        ],
    )
    assert_array_equal(
        return_path,
        [
            (1, 6),
            (1, 5),
            (1, 4),
            (1, 3),
            (1, 2),
            (1, 1),
            (2, 1),
            (3, 1),
            (4, 1),
            (5, 1),
            (6, 1),
            (7, 1),
            (7, 2),
        ],
    )


def test_offsets():
    offsets = [(1, i) for i in range(10)] + [(1, -i) for i in range(1, 10)]
    with expected_warnings(['Upgrading NumPy' + warning_optional]):
        m = mcp.MCP(a, offsets=offsets)
    costs, traceback = m.find_costs([(1, 6)])
    assert_array_equal(
        traceback,
        [
            [-2, -2, -2, -2, -2, -2, -2, -2],
            [-2, -2, -2, -2, -2, -2, -1, -2],
            [15, 14, 13, 12, 11, 10, 0, 1],
            [10, 0, 1, 2, 3, 4, 5, 6],
            [10, 0, 1, 2, 3, 4, 5, 6],
            [10, 0, 1, 2, 3, 4, 5, 6],
            [10, 0, 1, 2, 3, 4, 5, 6],
            [10, 0, 1, 2, 3, 4, 5, 6],
        ],
    )
    assert hasattr(m, "offsets")
    assert_array_equal(offsets, m.offsets)


@parametrize("shape", [(100, 100), (5, 8, 13, 17)] * 5)
def test_crashing(shape):
    _test_random(shape)


def _test_random(shape):
    # Just tests for crashing -- not for correctness.
    a = np.random.rand(*shape).astype(np.float32)
    starts = [
        [0] * len(shape),
        [-1] * len(shape),
        (np.random.rand(len(shape)) * shape).astype(int),
    ]
    ends = [(np.random.rand(len(shape)) * shape).astype(int) for i in range(4)]
    with expected_warnings(['Upgrading NumPy' + warning_optional]):
        m = mcp.MCP(a, fully_connected=True)
    costs, offsets = m.find_costs(starts)
    for point in [(np.random.rand(len(shape)) * shape).astype(int) for i in range(4)]:
        m.traceback(point)
    m._reset()
    m.find_costs(starts, ends)
    for end in ends:
        m.traceback(end)
    return a, costs, offsets



import numpy as np
import pytest

import skimage as ski


def test_blur_effect():
    """Test that the blur metric increases with more blurring."""
    image = ski.data.astronaut()
    B0 = ski.measure.blur_effect(image, channel_axis=-1)
    B1 = ski.measure.blur_effect(
        ski.filters.gaussian(image, sigma=1, channel_axis=-1), channel_axis=-1
    )
    B2 = ski.measure.blur_effect(
        ski.filters.gaussian(image, sigma=4, channel_axis=-1), channel_axis=-1
    )
    assert 0 <= B0 < 1
    assert B0 < B1 < B2


def test_blur_effect_h_size():
    """Test that the blur metric decreases with increasing size of the
    re-blurring filter.
    """
    image = ski.data.astronaut()
    B0 = ski.measure.blur_effect(image, h_size=3, channel_axis=-1)
    B1 = ski.measure.blur_effect(image, channel_axis=-1)  # default h_size is 11
    B2 = ski.measure.blur_effect(image, h_size=30, channel_axis=-1)
    assert 0 <= B0 < 1
    assert B0 > B1 > B2


def test_blur_effect_channel_axis():
    """Test that passing an RGB image is equivalent to passing its grayscale
    version.
    """
    image = ski.data.astronaut()
    B0 = ski.measure.blur_effect(image, channel_axis=-1)
    B1 = ski.measure.blur_effect(ski.color.rgb2gray(image))
    B0_arr = ski.measure.blur_effect(image, channel_axis=-1, reduce_func=None)
    B1_arr = ski.measure.blur_effect(ski.color.rgb2gray(image), reduce_func=None)
    assert 0 <= B0 < 1
    assert B0 == B1
    np.testing.assert_array_equal(B0_arr, B1_arr)


def test_blur_effect_3d():
    """Test that the blur metric works on a 3D image."""
    image_3d = ski.data.cells3d()[:, 1, :, :]  # grab just the nuclei
    B0 = ski.measure.blur_effect(image_3d)
    B1 = ski.measure.blur_effect(ski.filters.gaussian(image_3d, sigma=1))
    B2 = ski.measure.blur_effect(ski.filters.gaussian(image_3d, sigma=4))
    assert 0 <= B0 < 1
    assert B0 < B1 < B2


@pytest.mark.parametrize('factor', [0, 1, 2.5])
def test_blur_effect_uniform_input(factor):
    """Test that the blur metric is 1 for completely uniform images."""
    image = np.ones((10, 10, 3)) * factor
    B = ski.measure.blur_effect(image, channel_axis=-1)
    assert B == 1


@pytest.mark.parametrize('shape', [(10, 11), (4, 5, 6)])
def test_blur_single_axis_constant_image(shape):
    """Test that the blur metric is 1 for an image that is uniform along one axis."""
    row = np.linspace(0, 1, shape[-1])
    image = np.broadcast_to(row, shape)
    B = ski.measure.blur_effect(image)
    assert B == 1
